#pragma once

#include "libsyntax_global.h"
#include <libtoken.h>
#include <memory>
#include <tuple>
#include <functional>
namespace ast_basic {
	class ExprInst;
	class ExpressionElement;
}

namespace lib_syntax {
	class BaseRule;

	/**
	 * @brief ﷨쳣
	 */
	class  LIBSYNTAX_EXPORT SyntaxException {
	private:
		QString msg_store;

	public:
		SyntaxException(const QString& message);
		virtual ~SyntaxException() = default;

		virtual QString message() const;
	};

	// ﷨ӿ ===================================================================================================
	/**
	 * @brief Ľӿ
	 */
	class ParseContext {
	public:
		virtual ~ParseContext() = default;

		virtual void setCurrentFile(const QString &path) = 0;
		virtual QString currentFile() const = 0;

		virtual void appendParseErrors(const QString & file_path, int start, const QString &error_msg) = 0;
		virtual QStringList errors() const = 0;
		virtual void clearErrors(const QString &file_path, int start) = 0;

		virtual void appendDoc(std::shared_ptr<ast_basic::ExprInst> inst) = 0;
		virtual QList<std::shared_ptr<const ast_basic::ExprInst>> getDocs() const = 0;

		/**
		 * \brief ǰʽԪ.
		 *
		 * \return صǰʽ
		 */
		virtual std::shared_ptr<ast_basic::ExprInst> currentInst() const = 0;
		virtual void pushInst(std::shared_ptr<ast_basic::ExprInst> current_inst) = 0;
		virtual std::shared_ptr<ast_basic::ExprInst> popInst() = 0;

		virtual std::shared_ptr<const BaseRule> currentExpressionRule() const = 0;
		virtual void pushExpressionRule(std::shared_ptr<const BaseRule> inst) = 0;
		virtual std::shared_ptr<const BaseRule> popExpressionRule() = 0;
		virtual QList<std::shared_ptr<const BaseRule>> currentExpressionRuleStack() const = 0;
	};

	/**
	 * @brief ﷨ƥӿ
	 */
	class BaseRule {
	public:
		virtual ~BaseRule() = default;
		/**
		 * @brief ӹ
		 * @return
		 */
		virtual QList<std::shared_ptr<const BaseRule>> children() const = 0;

		/**
		 * @brief ƥ
		 */
		enum class MatchResult {
			Success, // ƥ
			Part,    // ƥ
			Fail     // ӵһȫƥ
		};

		/**
		 * @brief 
		 * @param rt_inst 
		 * @param head бͷ
		 * @return ؽ<ƥбͷ,ƥ䳤>
		 */
		virtual std::tuple<BaseRule::MatchResult, std::shared_ptr<const lib_token::IWordBase>>
				parse(std::shared_ptr<ParseContext> rt_inst, std::shared_ptr<const lib_token::IWordBase> head) const = 0;

		/**
		 * ƥ﷨Ĵʷб
		 *
		 * \return ʷ
		 */
		virtual QString token_present() const = 0;
	};


	/**
	 * @brief tokenƥ
	 */
	class LIBSYNTAX_EXPORT TokenMatch : public BaseRule, public std::enable_shared_from_this<TokenMatch> {
	private:
		std::shared_ptr<const lib_token::ITokenDefine> define_peer;

	public:
		TokenMatch(std::shared_ptr<const lib_token::ITokenDefine> define);

		// BaseRule interface
	public:
		virtual QList<std::shared_ptr<const BaseRule>> children() const override;
		virtual std::tuple<BaseRule::MatchResult, std::shared_ptr<const lib_token::IWordBase>>
				parse(std::shared_ptr<ParseContext> rt_inst, std::shared_ptr<const lib_token::IWordBase> head) const override;
		virtual QString token_present() const override;
	};

	/**
	 * @brief ﷨ƥ
	 */
	class LIBSYNTAX_EXPORT Any : public BaseRule, public std::enable_shared_from_this<Any> {
	private:
		QList<std::shared_ptr<const BaseRule>> mbrs_store;

		std::tuple<MatchResult, uint, std::shared_ptr<const BaseRule>, std::shared_ptr<const lib_token::IWordBase>> rule_select(std::shared_ptr<const lib_token::IWordBase> head) const;

	public:
		Any(const QList<std::shared_ptr<const BaseRule>> mbrs);

		// BaseRule interface
	public:
		virtual QList<std::shared_ptr<const BaseRule>> children() const override;
		virtual std::tuple<BaseRule::MatchResult, std::shared_ptr<const lib_token::IWordBase>>
				parse(std::shared_ptr<ParseContext> rt_inst, std::shared_ptr<const lib_token::IWordBase> head) const override;
		virtual QString token_present() const override;
	};

	/**
	 * @brief ﷨ƥ
	 */
	class LIBSYNTAX_EXPORT Seqs : public BaseRule, public std::enable_shared_from_this<Seqs> {
	private:
		QList<std::shared_ptr<const BaseRule>> mbrs_store;

	public:
		Seqs(const QList<std::shared_ptr<const BaseRule>> mbrs);

		// BaseRule interface
	public:
		virtual QList<std::shared_ptr<const BaseRule>> children() const override;
		virtual std::tuple<BaseRule::MatchResult, std::shared_ptr<const lib_token::IWordBase>>
				parse(std::shared_ptr<ParseContext> rt_inst, std::shared_ptr<const lib_token::IWordBase> head) const override;
		virtual QString token_present() const override;
	};

	/**
	 * @brief ﷨ظƥ
	 */
	class LIBSYNTAX_EXPORT Rept : public BaseRule, public std::enable_shared_from_this<Rept> {
	private:
		std::shared_ptr<const BaseRule> rule_peer;
		int min_match, max_match;

	public:
		Rept(std::shared_ptr<const BaseRule> rule, int min, int max);

		// BaseRule interface
	public:
		virtual QList<std::shared_ptr<const BaseRule>> children() const override;
		virtual std::tuple<BaseRule::MatchResult, std::shared_ptr<const lib_token::IWordBase>>
				parse(std::shared_ptr<ParseContext> rt_inst, std::shared_ptr<const lib_token::IWordBase> head) const override;
		virtual QString token_present() const override;
	};

	// ﷨ʵ ===================================================================================================
	/**
	 * @brief Ӧ﷨ʽ
	 */
	class LIBSYNTAX_EXPORT ExprRule : public lib_syntax::BaseRule, public std::enable_shared_from_this<ExprRule> {
	public:
		typedef QList<std::shared_ptr<const lib_token::IToken>> TokenSeqs;
		ExprRule(const QString& rule_name, int expr_mark);

		virtual std::shared_ptr<const ExprRule> reloadRule(std::shared_ptr<const BaseRule> rule);
		virtual QString name() const;
		virtual int typeMark() const;

		virtual std::shared_ptr<ast_basic::ExprInst> newEmptyInstance() const = 0;
		virtual std::shared_ptr<ExprRule> makeCopy() const = 0;

		// BaseRule interface
	public:
		virtual QList<std::shared_ptr<const lib_syntax::BaseRule>> children() const override;
		virtual std::tuple<BaseRule::MatchResult, std::shared_ptr<const lib_token::IWordBase>>
				parse(std::shared_ptr<ParseContext> rt_inst, std::shared_ptr<const lib_token::IWordBase> head) const override;
		virtual QString token_present() const override;

	private:
		std::shared_ptr<const lib_syntax::BaseRule> child_store;
		QString name_store;
		int mark_store;
	};

	/**
	 * ﷨Ԫؽ.
	 */
	template<class ExprType>
	class ElementRule : public ExprRule {
	public:
		ElementRule(const QString& rule_name, int expr_mark)
			:ExprRule(rule_name, expr_mark){}

		virtual std::shared_ptr<ast_basic::ExprInst> newEmptyInstance() const {
			return std::dynamic_pointer_cast<ast_basic::ExprInst>(
				std::make_shared<ExprType>(this->shared_from_this()));
		}

		virtual std::shared_ptr<ExprRule> makeCopy() const {
			return std::make_shared<ElementRule<ExprType>>(name(), typeMark());
		}
	};
} // namespace lib_syntax